#include "globals.h"
#include "skimic.h"
#include "editsess.h"
#include "skkimmgr.h"

class CSyncEditSession : public CEditSessionBase 
{
public:
    CSyncEditSession (ITfContext *pContext, BOOL bUpdateCMode, CSkkImeMgr* pSkkIme) :
		CEditSessionBase(pContext), 
		m_bUpdateCMode (bUpdateCMode), 
		m_pSkkIme (pSkkIme)
	{
		return ;
	}
	
    // ITfEditSession
	STDMETHODIMP DoEditSession (TfEditCookie ec) {
		return	m_pSkkIme->OnRestoreEditSession (m_pContext, ec, m_bUpdateCMode) ;
	}
	
private:
	BOOL		m_bUpdateCMode ;
	CSkkImeMgr*	m_pSkkIme ;
} ;

class CCompositeClearSession : public CEditSessionBase {
public:
	CCompositeClearSession (ITfContext* pContext, CSkkImeTextService* pTSF) : CEditSessionBase (pContext), m_pTSF (pTSF) {
		return ;
	}
	STDMETHODIMP	DoEditSession (TfEditCookie ec) ;

private:
	CSkkImeTextService*		m_pTSF ;
} ;

class CToggleImeEditSession : public CEditSessionBase {
public:
	CToggleImeEditSession (ITfContext* pContext, CSkkImeMgr* pSkkIme) : CEditSessionBase (pContext), m_pSkkIme (pSkkIme) {
		return ;
	}
	STDMETHODIMP	DoEditSession (TfEditCookie ec) ;
private:
	CSkkImeMgr*	m_pSkkIme ;
} ;

STDAPI
CCompositeClearSession::DoEditSession (
	TfEditCookie		ec)
{
    TF_SELECTION		tfSelection ;
	ULONG				cFetched ;
	ITfComposition*		pComposition	= NULL ;
	ITfRange*			pRange		= NULL ;
	ITfProperty*		pProp 		= NULL ;
	ITfProperty*		pPropAttr	= NULL ;
	BOOL				fComposing	= FALSE ;

    if (m_pContext->GetSelection (ec, TF_DEFAULT_SELECTION, 1, &tfSelection, &cFetched) != S_OK ||
		cFetched != 1)
		return	S_OK ;

	fComposing	= m_pTSF->_IsComposing (&pComposition) ;
	if (! fComposing) 
		goto	Exit ;

	if (pComposition == NULL || pComposition->GetRange (&pRange) != S_OK || pRange == NULL) 
		goto	Exit ;

    if (SUCCEEDED (m_pContext->GetProperty (GUID_PROP_ATTRIBUTE, &pPropAttr) && pPropAttr != NULL)) {
		pPropAttr->Clear (ec, pRange) ;
		pPropAttr->Release () ;
	}

#if defined (buggy_code_using_prop_reading) || 1
	if (SUCCEEDED (m_pContext->GetProperty (GUID_PROP_READING, &pProp) && pProp != NULL)) {
		pProp->Clear (ec, pRange) ;
	} else {
		pProp	= NULL ;
	}
#endif
	if (pRange->SetText (ec, 0, L"", 0) != S_OK) 
		goto	Exit ;
	if (pComposition->ShiftStart (ec, pRange) != S_OK ||
		pComposition->ShiftEnd   (ec, pRange) != S_OK)
		goto	Exit ;
	tfSelection.range->ShiftStartToRange (ec, pRange, TF_ANCHOR_START) ;
	tfSelection.range->Collapse (ec, TF_ANCHOR_START) ;
	m_pContext->SetSelection (ec, 1, &tfSelection) ;

  Exit:
	if (pProp != NULL)
		pProp->Release () ;
	if (pRange != NULL)
		pRange->Release () ;
	tfSelection.range->Release () ;
	return	S_OK ;
}

STDAPI
CToggleImeEditSession::DoEditSession (
	TfEditCookie		ec)
{
	HRESULT	hr ;

	DEBUGPRINTF ((TEXT ("CToggleImeEditSession::DoEditSession ()\n"))) ;

	hr	= m_pSkkIme->OnToggleImeSession (m_pContext, ec) ;
	if (FAILED (hr)) {
		DEBUGPRINTF ((TEXT ("m_pSkkIme->OnToggleImeSession failed.\n"))) ;
		return	hr ;
	}
	return	S_OK ;
}

HRESULT
CSkkImeTextService::_ToggleIME (
	ITfContext*		pContext,
	HRESULT*		phrEaten)
{
    CToggleImeEditSession*	pEditSession ;
	HRESULT	hr, hrEaten ;

	pEditSession = new CToggleImeEditSession (pContext, m_pSkkIme) ;
	if (pEditSession == NULL) {
		*phrEaten	= S_OK ;
		return	E_OUTOFMEMORY ;
	}	
	hr	= pContext->RequestEditSession(m_tfClientId, pEditSession, TF_ES_SYNC | TF_ES_READWRITE, &hrEaten) ;
	if (FAILED (hr))
		hrEaten	= S_OK ;
    pEditSession->Release () ;
	*phrEaten	= hrEaten ;

	if (_IsKeyboardOpen ()) {
		_UnregisterImeOnKeys () ;
		//_RegisterImeOffKeys () ;
	} else {
		_RegisterImeOnKeys () ;
		//_UnregisterImeOffKeys () ;
	}
	return	hr ;
}

HRESULT
CSkkImeTextService::_SyncComposition (
	ITfContext*		pContext,
	BOOL			fToggle,
	BOOL			fCMode)
{
	CSyncEditSession*	pEditSession	= NULL ;
	HRESULT	hr	= S_OK ;
	HRESULT	hrDummy ;

	if (fToggle) {
		CToggleImeEditSession*	pEditSession	= NULL ;

		pEditSession = new CToggleImeEditSession (pContext, m_pSkkIme) ;
		if (pEditSession == NULL) 
			return	E_OUTOFMEMORY ;
		
		hr	= pContext->RequestEditSession(m_tfClientId, pEditSession, TF_ES_SYNC | TF_ES_READWRITE, &hrDummy) ;
		pEditSession->Release () ;
		goto	Exit ;
	} 
	if (m_pSkkIme->NeedClearCompositionp ()) {
		CCompositeClearSession*	pClearSession	= NULL ;
		/*	OutlookExpress̏ SetText (L" ",1) ɂ1pCompositionText̏
		 *	s\Ȃ̂ŁAOɃeLXgNAB
		 *	Ƃ΁Aȃ[h d,space ̏Ԃɓ͂ƌʂ "d " ɂȂ邱ƂŁA̖
		 *	͍ČłB({ d  " " ɂȂ)
		 */
		pClearSession	= new CCompositeClearSession (pContext, this) ;
		if (pClearSession == NULL) {
			hr	= E_OUTOFMEMORY ;
			goto	Exit ;
		}
		/*	SYNCƁAPɂĂ܂悤ŁA܊pClear𕪂ӖȂȂB
		 *	SYNC͍B
		 */
		(void) pContext->RequestEditSession (m_tfClientId, pClearSession, TF_ES_READWRITE, &hrDummy) ;
		/*	̕Ԃl͂ǂׂH
		 */
		pClearSession->Release () ;
	}

	/*	ۂɕ̃L[Xg[N̂ꏊɁuŁv
	 *	}KvB
	 *	ǂ̕邩́A̎̎BȂA
	 *	IƃL[Xg[N͏ƂɂȂ̂ȁB
	 *
	 *	̑} CKeystrokeEditSession::DoEditSession 
	 *	ǊB
	 */
	pEditSession = new CSyncEditSession (pContext, fCMode, m_pSkkIme) ;
	if (pEditSession == NULL) {
		hr	= E_OUTOFMEMORY ;
		goto	Exit ;
	}
	
	/*	u}vȂǂ̏s߂ɂ̓bNK
	 *	v(炵)B=> we need a lock to do our work
	 *
	 *	nb:  method  TF_ES_SYNC tOgƂ legal
	 *		ށAH@Hȏꏊ1łB
	 *	ȂA SYNC AƂƂB
	 */
	(void) pContext->RequestEditSession(m_tfClientId, pEditSession, TF_ES_READWRITE, &hrDummy) ;
	pEditSession->Release () ;
Exit:
	return hr ;
}

