#if defined (UNITTEST)
#include "StdAfx.h"
#else
#include "local.h"
#endif
#include "ImeDocP.h"
#include "ImeBuffer.h"
#include "TMarker.h"
#include "lmstate.h"
#include "TCutBufferSession.h"

/*======================================================================== kill-ring-save
 */
static	int		LM_bKillRingSave_Exit (struct CImeDoc*) ;

int
LM_bKillRingSave (
	struct CImeDoc*		pDoc)
{
	if (! ImeDoc_bCall (pDoc, LM_bKillRegion, &LM_bKillRingSave_Exit))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bKillRingSave_Exit (
	struct CImeDoc*		pDoc)
{
	ImeDoc_vJump (pDoc, LM_bYank) ;
	return	LMR_CONTINUE ;
}

/*======================================================================== kill-region
 */
int
LM_bKillRegion (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pCurBuffer ;
	struct TMarker*		pmkPoint ;
	struct TMarker*		pmkEnd ;
	struct TMarker*		pmkTop ;
	struct TMarker*		pmkMark ;
	int			nLastCmd, nBufferTop, nBufferEnd, nStart, nEnd ;
	BOOL		fAppend ;
	DCHAR		rszBuffer [MAXCOMPLEN] ;
	LPCDSTR		pdText ;
	int			nText ;
	LPCDSTR		pSrc ;
	LPDSTR		pDest ;
	int			nSrc, nDest ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pCurBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pCurBuffer == NULL)
		return	LMR_ERROR ;

	if (! ImeBuffer_bGetMarker (pCurBuffer, MARKER_POINT,     &pmkPoint) || pmkPoint == NULL ||
		! ImeBuffer_bGetMarker (pCurBuffer, MARKER_BUFFEREND, &pmkEnd)   || pmkEnd   == NULL ||
		! ImeBuffer_bGetMarker (pCurBuffer, MARKER_MARK,      &pmkMark)  || pmkMark  == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (ImeBuffer_bGetMarker (pCurBuffer, MARKER_BUFFERTOP, &pmkTop) && pmkTop != NULL) {
		nBufferTop	= TMarker_iGetPosition (pmkTop) ;
	} else {
		nBufferTop	= 0 ;
	}
	nStart			= TMarker_iGetPosition (pmkPoint) ;
	nEnd			= TMarker_iGetPosition (pmkMark) ;
	if (nStart > nEnd) {
		int	nTmp ;
		nTmp	= nStart ;
		nStart	= nEnd ;
		nEnd	= nTmp ;
	}

	nBufferEnd		= TMarker_iGetPosition (pmkEnd) ;
	if (nStart == nEnd || nStart < nBufferTop || nEnd > nBufferEnd) {
		return	LMR_RETURN ;
	}

	nLastCmd	= ImeDoc_iGetLastCommand (pDoc) ;
	fAppend		= (nLastCmd == NFUNC_KILL_REGION)? TRUE : FALSE ;

	/*	ŊOEɉsR[h \n  0x0D, 0x0A
	 *	ɕϊKvB
	 */
	pdText	= ImeBuffer_pBufferRawString (pCurBuffer, &nText) ;
	if (nStart > nText || nEnd > nText) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	pSrc	= pdText + nStart ;
	nSrc	= nEnd - nStart ;
	pDest	= rszBuffer ;
	nDest	= ARRAYSIZE (rszBuffer) ;

	while (nSrc > 0 && nDest > 0) {
		if (*pSrc == L'\n') {
			/*	0x0D, 0x0A Ƒ̂łȂ
			 *	1ȂƂɂB*/
			if (nDest > 1) {
				*pDest ++	= 0x0D ;					
				*pDest ++	= 0x0A ;
				nDest  -= 2 ;
			}
		} else {
			*pDest ++	= *pSrc ;
			nDest -- ;
		}
		pSrc ++ ;
		nSrc -- ;
	}
	if (! TCutBufferSession_bSet (rszBuffer, pDest - rszBuffer, fAppend)) {
		/* ... */
	}
	if (! ImeBuffer_bDeleteRegion (pCurBuffer, nStart, nEnd)) {
	}
	ImeDoc_bSetThisCommand (pDoc, NFUNC_KILL_REGION) ;
	return	LMR_RETURN ;
}

/*======================================================================== yank
 */
int
LM_bYank (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pCurBuffer ;
	DCHAR		rszBuffer [MAXCOMPLEN] ;
	LPCDSTR		pdSrc, pdSrcBase ;
	int			n ;

	/*	signal ݒ肳ĂΓ삵ȂB
	 *	ƂA̔𖈉Ă΂ꂽɏ̂͂ȂcB
	 */
	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pCurBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pCurBuffer == NULL)
		return	LMR_ERROR ;

	n	= TCutBufferSession_iGet (rszBuffer, ARRAYSIZE (rszBuffer)) ;
	if (n > 0) {
		pdSrc	= rszBuffer ;
		while (n > 0) {
			pdSrcBase	= pdSrc ;
			while (n > 0 && *pdSrc != 0x0D) {
				pdSrc	++ ;
				n		-- ;
			}
			if (! ImeBuffer_bInsertAndInherit (pCurBuffer, pdSrcBase, pdSrc - pdSrcBase)) {
				ImeDoc_vSetSignalError (pDoc) ;
				return	LMR_RETURN ;
			}

			/*	OEƂ̃C^[tF[Xł邱̕ŉsR[h
			 *	0x0D, 0x0A ƂĂė\B
			 */
			if (*pdSrc == 0x0D) {
				if (n > 1 && *(pdSrc + 1) == 0x0A) {
					while (n > 1 && *pdSrc == 0x0D && *(pdSrc + 1) == 0x0A) {
						if (! ImeBuffer_bInsertAndInheritW (pCurBuffer, L"\n", 1)) {
							ImeDoc_vSetSignalError (pDoc) ;
							return	LMR_RETURN ;
						}
						pdSrc	+= 2 ;
						n		-= 2 ;
					}
				} else {
					/*	0x0D ̏o͂̂܂ܒʂB
					 */
					if (! ImeBuffer_bInsertAndInherit (pCurBuffer, pdSrc, 1)) {
						ImeDoc_vSetSignalError (pDoc) ;
						return	LMR_RETURN ;
					}
					pdSrc	++ ;
					n		-- ;
				}
			}
		}
	}
	return	LMR_RETURN ;
}
